import numpy as np
from evola.solution import Solution

# Optionally define random seed
#np.random.seed(42)


class EvolutiveParticle(Solution):
    def __init__(self, chromossome, cost_function, cost_function_args, WI: float, WM: float, WC: float):
        super().__init__(chromossome, cost_function, cost_function_args)
        self.WI = WI
        self.WM = WM
        self.WC = WC

        self.calc_cost()

        return

    def mutate(self):
        self.WI = self.WI * (1 + 0.1 * np.random.normal())
        self.WM = self.WM * (1 + 0.1 * np.random.normal())
        self.WC = self.WC * (1 + 0.1 * np.random.normal())



import csv
import sys
from copy import deepcopy
from itertools import islice
from operator import attrgetter
from typing import List

import numpy as np

from evola.epso.particle import EvolutiveParticle
from evola.scene import Scene


class EvolutiveSwarm:
    def __init__(
        self,
        swarm_size: int,
        WI: float,
        WM: float,
        WC: float,
        communication_p: float,
        scene: Scene,
    ):
        self.size = swarm_size
        self.gen = 0  # generation counter (for WI)
        self.scene = scene
        self._communication_p = communication_p  # communication probability
        ancestors: List[EvolutiveParticle] = []
        particles: List[EvolutiveParticle] = []

        # Generate particles and ancestors randomly
        for _ in range(swarm_size):
            chromossome1 = np.zeros(self.scene.chromossome_length)
            chromossome2 = np.zeros(self.scene.chromossome_length)
            for i in range(self.scene.chromossome_length):
                chromossome1[i] = np.random.uniform(low=self.scene.chromossome_low, high=self.scene.chromossome_high)
                chromossome2[i] = np.random.uniform(low=self.scene.chromossome_low, high=self.scene.chromossome_high)

            ancestors.append(
                EvolutiveParticle(
                    chromossome1,
                    scene.cost_function,
                    scene.cost_function_args,
                    WI,
                    WM,
                    WC,
                )
            )
            particles.append(
                EvolutiveParticle(
                    chromossome2,
                    scene.cost_function,
                    scene.cost_function_args,
                    WI,
                    WM,
                    WC,
                )
            )

        self.ancestors: List[EvolutiveParticle] = ancestors
        self.best_ancestors: List[EvolutiveParticle] = deepcopy(ancestors)
        self.particles: List[EvolutiveParticle] = particles

        sorted = min(self.ancestors, key=attrgetter("cost"))
        self.global_best = deepcopy(sorted)  # best particle

        return

    def reproduce(self):

        # Append a copy of the swarm to itself
        sons = deepcopy(self.particles)
        son: EvolutiveParticle
        for son in sons:
            son.mutate()

        self.particles.extend(sons)
        self.ancestors.extend(self.ancestors)
        self.best_ancestors.extend(self.best_ancestors)

        # Swarm doubles it's size
        self.size = self.size * 2

        return

    def select(self):

        # Population reduces in half
        self.size = int(self.size / 2)

        # Overwrite parents, ancestors and best ancestors if children perform better after swarm moves
        for i in range(self.size):
            if self.particles[i].cost > self.particles[i + self.size].cost:
                self.particles[i] = deepcopy(self.particles[i + self.size])
                self.ancestors[i] = deepcopy(self.ancestors[i + self.size])
                self.best_ancestors[i] = deepcopy(self.best_ancestors[i + self.size])

        # Only first half of the swarm survives
        self.particles = self.particles[: self.size]
        self.ancestors = self.ancestors[: self.size]
        self.best_ancestors = self.best_ancestors[: self.size]

        return

    def move(self):

        self.gen = self.gen + 1  # Increment nº of generations

        # Create new generation
        new_particles = []

        # Move swarm
        for i in range(self.size):
            chromossome = np.zeros(len(self.particles[i].chromossome))

            # Create probability of alllowing the particle to move on each dimension of the global best
            if self._communication_p == 1:
                communication_matrix = np.ones(len(self.particles[i].chromossome))
            else:
                communication_matrix = np.random.choice(
                    [0, 1],
                    len(self.particles[i].chromossome),
                    p=[1 - self._communication_p, self._communication_p],
                )

            # Apply deviation, one dimention at a time
            for j in range(len(self.particles[i].chromossome)):
                deviation = (
                    1
                    / self.gen
                    * self.particles[i].WI
                    * (self.particles[i].chromossome[j] - self.ancestors[i].chromossome[j])
                )
                deviation = deviation + np.random.normal() * self.particles[i].WM * (
                    self.best_ancestors[i].chromossome[j] - self.particles[i].chromossome[j]
                )
                deviation = deviation + communication_matrix[j] * np.random.normal() * self.particles[i].WC * (
                    self.global_best.chromossome[j] - self.particles[i].chromossome[j]
                )
                chromossome[j] = self.particles[i].chromossome[j] + deviation

            # Apply chromossome value ceiling and floor
            chromossome[chromossome > self.scene.chromossome_high] = self.scene.chromossome_high
            chromossome[chromossome < self.scene.chromossome_low] = self.scene.chromossome_low

            new_particles.append(
                EvolutiveParticle(
                    chromossome,
                    self.scene.cost_function,
                    self.scene.cost_function_args,
                    self.particles[i].WI,
                    self.particles[i].WM,
                    self.particles[i].WC,
                )
            )

        # Current generation is now ancestor generation
        self.ancestors = deepcopy(self.particles)

        # Update best ancestors according to new generation of ancestors
        for i in range(len(self.best_ancestors)):
            if self.ancestors[i].cost < self.best_ancestors[i].cost:
                self.best_ancestors[i] = deepcopy(self.ancestors[i])

        # Update new generation particles
        self.particles = new_particles

        # Update global best
        best = min(self.ancestors, key=attrgetter("cost"))
        self.global_best = deepcopy(best)

        return

    def last_gen_update(self):
        # Must run after last generation
        newbest = min(self.particles, key=attrgetter("cost"))
        if newbest.cost < self.global_best.cost:
            self.global_best = newbest
        return

    def average_cost(self):
        sum = 0.0
        p: EvolutiveParticle
        for p in self.particles:
            sum = sum + p.cost

        return sum / self.size

    def display(self, top):
        particles = sorted(self.particles, key=attrgetter("cost"))
        print("\n")
        for i in range(top):
            print("Particle %d :" % i)
            print("Cost: %.3f" % particles[i].cost)
            print("Chromossome values")
            print(particles[i].chromossome)
            print("\n")

        return

    def export(self, top, filename):
        particles = sorted(self.particles, key=attrgetter("cost"))
        with open(sys.path[0] + "/" + filename, "w", newline="") as csvfile:
            writer = csv.writer(csvfile, delimiter=" ")

            header = ["Solutions"]
            header.extend(str(list(range(1, top + 1))))

            writer.writerow(header)
            writer.writerow(["Chromossome value"])
            for i in range(self.scene.chromossome_length):
                chromossomes = []
                for particle in islice(particles, 0, top):
                    chromossomes.append(particle.chromossome[i])
                writer.writerow(["T%d" % (i + 1)] + chromossomes)
            costs = ["Cost"]
            for particle in islice(particles, 0, top):
                costs.append(str(particle.cost))
            writer.writerow(costs)

        return



from copy import deepcopy
from multiprocessing import Queue
from tqdm.auto import tqdm
import time

from evola.epso.swarm import EvolutiveSwarm
from evola.scene import Scene
from evola.simulation import Simulation


class EPSO(Simulation):
    def __init__(self,generations,size: int,WI:float,WM:float,WC:float,communication_probability:float,scene: Scene, hold: bool, export_top=0) -> None:
        desc = scene.desc + " EPSO (p=%.2f)"%communication_probability
        super().__init__(generations,EvolutiveSwarm(size,WI,WM,WC,communication_probability,scene),hold,desc,export_top)

        self.pop:EvolutiveSwarm
        return

    @property
    def best_particle(self):
        return self.pop.global_best
    
    def run_gui(self):
        self.init_graph()
        history = []
        start_time = time.time()

        for _ in tqdm(range(self.generations)):
            # EPSO steps
            self.pop.reproduce()
            self.pop.move()
            self.pop.select()

            # Obter o custo melhor da população
            new_data = self.pop.global_best.cost
            history.append(new_data)
            self.cost_history.append(new_data)
            # Atualizar gráfico
            self.update_graph(history)

        self.pop.last_gen_update()

        end_time = time.time()

        self.best_cost = self.pop.global_best.cost

        print(self.final_message(start_time,end_time))
        
        if self.export_top>0:
            self.export(self.export_top,self.generations)

        self.finish_graph()

        return

    def run_cli(self,q:Queue=None,thread_num=0,verbose=True,itera=1):

        start_time = time.time()
        
        if verbose and itera==1:
            pbar = tqdm(range(self.generations),position=thread_num+1, leave=False)
            pop = deepcopy(self.pop)
            for _ in pbar:
                # EPSO
                self.cost_history.append(deepcopy(pop.global_best.cost))
                pbar.set_description(f'({self.pop.size},{self.generations}) C = {pop.global_best.cost:.0f} euros')
                pop.reproduce()
                pop.move()
                pop.select()
            pop.last_gen_update()
                

        elif verbose and itera > 1:
            pbar = tqdm(range(itera),position=thread_num+1, leave=False)
            for _ in pbar:
                pop = deepcopy(self.pop)
                for _ in range(self.generations):
                    # EPSO
                    pop.reproduce()
                    pop.move()
                    pop.select()
                pop.last_gen_update()
                self.best_hist.append(deepcopy(pop.global_best.cost))
                self.avg_hist.append(pop.average_cost())
        else:
            for _ in range(itera):
                pop = deepcopy(self.pop)
                for _ in range(self.generations):
                    # EPSO
                    pop.reproduce()
                    pop.move()
                    pop.select()
                pop.last_gen_update()
                self.best_hist.append(deepcopy(pop.global_best.cost))
                self.avg_hist.append(pop.average_cost())
        
        
        end_time = time.time()

        if itera==1:
            self.best_cost = pop.global_best.cost
        else:
            self.best_cost = min(self.best_hist)
        if verbose:
            print(self.final_message(start_time,end_time))
        self.pop = pop

        if self.export_top>0:
            self.export(self.export_top)
        if q:
            q.put(self) # returns self in a queue, if multiprocessing is enabled
        return



from typing import Callable

class Solution:
    def __init__(self, chromossome, cost_function: Callable, cost_function_args: tuple):
        self.chromossome = chromossome
        self.cost: float = 0.0
        self._cost_function = cost_function
        self._cost_function_args = cost_function_args
        return

    def calc_cost(self):
        self.cost = self._cost_function(self.chromossome, *self._cost_function_args)


import csv
import sys
from matplotlib import pyplot as plt


class Simulation:
    def __init__(self, gen, pop, hold: bool, desc, export_top=0) -> None:
        self.generations = gen
        self.pop = pop
        self.hold = hold
        self.desc = desc + " (%d ind, %d gen)\n" % (self.pop.size, self.generations)
        self.export_top = export_top
        self.cost_history = []
        self.best_hist = []
        self.avg_hist = []
        self.best_cost = 0.0

        return

    def init_graph(self):
        # Inicialização do gráfico em tempo real
        plt.ion()
        fig,ax1 = plt.subplots()
        plt.ylabel('Cost')
        plt.xlabel('Generation')
        line1, = ax1.plot([], 'r-') 
        self._fig = fig
        self._ax = ax1
        self._line = line1


    def update_graph(self,history):
        new_data = history[-1]
        # Update y scale dynamically
        if len(history)==1: 
            gap = new_data * 0.5
            self._ax.set_ylim([new_data-gap, new_data+gap])
        else: 
            oldlim = self._ax.get_ylim()
            if new_data > oldlim[1]:
                tol = (new_data - oldlim[0]) * 0.1
                self._ax.set_ylim([oldlim[0], new_data + tol])
            elif new_data < oldlim[0]:
                tol = (oldlim[1] - new_data) * 0.1
                self._ax.set_ylim([new_data - tol, oldlim[1]])

        # Mudar eixo X
        self._line.set_xdata([*range(len(history))])
        self._ax.set_xlim(0,len(history))

        # Atualizar eixo Y
        self._line.set_ydata(history)

        # Atualizar gráfico
        self._fig.canvas.draw()
        self._fig.canvas.flush_events()
        return

    def finish_graph(self):
        self._fig.canvas.draw()
        plt.ioff()
        if self.hold:
            plt.show()  
        else:
            plt.close(self._fig)

    def export(self,top):
        name = self.desc + " (%d ind, %d gen)" % (self.pop.size, self.generations) + ".csv"
        self.pop.export(top=top,filename=name)
        return

    def final_message(self, start,end):
        head = self.desc 
        time = "\t-> Simulation time: %.3fs total | %.1f gen/s\n" % (end-start, self.generations/(end-start))
        bestcost =  "\t-> Least cost solution: %.6f\n" %  self.best_cost
        return head + time + bestcost
 
    def print(self, top):
        self.pop.display(top)
        return

    def export_histogram(self):
        if len(self.best_hist) > 0:
            filename = self.desc

            with open(sys.path[0] + "/" + filename + ".csv", 'w',newline='') as csvfile:
                writer = csv.writer(csvfile, delimiter=' ')
                for value in self.best_hist:
                    writer.writerow([value])

        return


from typing import Callable


class Scene:
    def __init__(
        self,
        chromossome_length: int,
        cost_function: Callable,
        cost_function_args: tuple,
        desc: str,
        chromossome_low: float = 0,
        chromossome_high: float = 1,
    ) -> None:
        self.desc = desc
        self.chromossome_length = chromossome_length
        self.chromossome_low = chromossome_low
        self.chromossome_high = chromossome_high
        self.cost_function = cost_function
        self.cost_function_args = cost_function_args



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['evola', 'evola.epso']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.6.0,<4.0.0', 'tqdm>=4.64.1,<5.0.0']

setup_kwargs = {
    'name': 'evola',
    'version': '0.0.1',
    'description': 'Evolutionary algorithms',
    'long_description': None,
    'author': 'David Freire',
    'author_email': 'up201806711@up.pt',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.9',
}


setup(**setup_kwargs)
